__title__ = "videohash2"
__description__ = (
    "Python package for Near Duplicate Video Detection "
    + "(Perceptual Video Hashing) - Get a 64-bit comparable "
    + "hash-value for any video."
)

__url__ = "https://demmenie.github.io/videohash2/"
__version__ = "3.1.0"
__status__ = "production"
__author__ = "Akash Mahanty and Chico Demmenie"
__author_email__ = "cdemmenie@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright 2024 Akash Mahanty"
